@@ -1,16 +1,17 @@
 import logging
 import warnings
 import threading
+from banal import ensure_list
 
+from sqlalchemy import func, select, false
 from sqlalchemy.sql import and_, expression
 from sqlalchemy.sql.expression import bindparam, ClauseElement
 from sqlalchemy.schema import Column, Index
-from sqlalchemy import func, select, false
 from sqlalchemy.schema import Table as SQLATable
 from sqlalchemy.exc import NoSuchTableError
 
-from dataset.types import Types
-from dataset.util import index_name, ensure_tuple
+from dataset.types import Types, MYSQL_LENGTH_TYPES
+from dataset.util import index_name
 from dataset.util import DatasetException, ResultIter, QUERY_STEP
 from dataset.util import normalize_table_name, pad_chunk_columns
 from dataset.util import normalize_column_name, normalize_column_key
@@ -21,20 +22,27 @@ log = logging.getLogger(__name__)
 
 class Table(object):
     """Represents a table in a database and exposes common operations."""
-    PRIMARY_DEFAULT = 'id'
 
-    def __init__(self, database, table_name, primary_id=None,
-                 primary_type=None, auto_create=False):
+    PRIMARY_DEFAULT = "id"
+
+    def __init__(
+        self,
+        database,
+        table_name,
+        primary_id=None,
+        primary_type=None,
+        auto_create=False,
+    ):
         """Initialise the table from database schema."""
         self.db = database
         self.name = normalize_table_name(table_name)
         self._table = None
         self._columns = None
         self._indexes = []
-        self._primary_id = primary_id if primary_id is not None \
-            else self.PRIMARY_DEFAULT
-        self._primary_type = primary_type if primary_type is not None \
-            else Types.integer
+        self._primary_id = (
+            primary_id if primary_id is not None else self.PRIMARY_DEFAULT
+        )
+        self._primary_type = primary_type if primary_type is not None else Types.integer
         self._auto_create = auto_create
 
     @property
@@ -69,10 +77,6 @@ class Table(object):
                     self._columns[key] = name
             return self._columns
 
-    def _flush_metadata(self):
-        with self.db.lock:
-            self._columns = None
-
     @property
     def columns(self):
         """Get a listing of all columns that exist in the table."""
@@ -205,8 +209,7 @@ class Table(object):
         if return_count:
             return self.count(clause)
 
-    def update_many(self, rows, keys, chunk_size=1000, ensure=None,
-                    types=None):
+    def update_many(self, rows, keys, chunk_size=1000, ensure=None, types=None):
         """Update many rows in the table at a time.
 
         This is significantly faster than updating them one by one. Per default
@@ -216,8 +219,7 @@ class Table(object):
         See :py:meth:`update() <dataset.Table.update>` for details on
         the other parameters.
         """
-        # Convert keys to a list if not a list or tuple.
-        keys = keys if type(keys) in (list, tuple) else [keys]
+        keys = ensure_list(keys)
 
         chunk = []
         columns = []
@@ -229,16 +231,14 @@ class Table(object):
 
             # bindparam requires names to not conflict (cannot be "id" for id)
             for key in keys:
-                row['_%s' % key] = row[key]
+                row["_%s" % key] = row[key]
 
             # Update when chunk_size is fulfilled or this is the last row
             if len(chunk) == chunk_size or index == len(rows) - 1:
-                cl = [self.table.c[k] == bindparam('_%s' % k) for k in keys]
+                cl = [self.table.c[k] == bindparam("_%s" % k) for k in keys]
                 stmt = self.table.update(
                     whereclause=and_(*cl),
-                    values={
-                        col: bindparam(col, required=False) for col in columns
-                    }
+                    values={col: bindparam(col, required=False) for col in columns},
                 )
                 self.db.executable.execute(stmt, chunk)
                 chunk = []
@@ -261,8 +261,7 @@ class Table(object):
             return self.insert(row, ensure=False)
         return True
 
-    def upsert_many(self, rows, keys, chunk_size=1000, ensure=None,
-                    types=None):
+    def upsert_many(self, rows, keys, chunk_size=1000, ensure=None, types=None):
         """
         Sorts multiple input rows into upserts and inserts. Inserts are passed
         to insert_many and upserts are updated.
@@ -270,8 +269,7 @@ class Table(object):
         See :py:meth:`upsert() <dataset.Table.upsert>` and
         :py:meth:`insert_many() <dataset.Table.insert_many>`.
         """
-        # Convert keys to a list if not a list or tuple.
-        keys = keys if type(keys) in (list, tuple) else [keys]
+        keys = ensure_list(keys)
 
         to_insert = []
         to_update = []
@@ -310,25 +308,25 @@ class Table(object):
     def _reflect_table(self):
         """Load the tables definition from the database."""
         with self.db.lock:
-            self._flush_metadata()
+            self._columns = None
             try:
-                self._table = SQLATable(self.name,
-                                        self.db.metadata,
-                                        schema=self.db.schema,
-                                        autoload=True)
+                self._table = SQLATable(
+                    self.name, self.db.metadata, schema=self.db.schema, autoload=True
+                )
             except NoSuchTableError:
                 self._table = None
 
     def _threading_warn(self):
         if self.db.in_transaction and threading.active_count() > 1:
-            warnings.warn("Changing the database schema inside a transaction "
-                          "in a multi-threaded environment is likely to lead "
-                          "to race conditions and synchronization issues.",
-                          RuntimeWarning)
+            warnings.warn(
+                "Changing the database schema inside a transaction "
+                "in a multi-threaded environment is likely to lead "
+                "to race conditions and synchronization issues.",
+                RuntimeWarning,
+            )
 
     def _sync_table(self, columns):
         """Lazy load, create or adapt the table structure in the database."""
-        self._flush_metadata()
         if self._table is None:
             # Load an existing table from the database.
             self._reflect_table()
@@ -339,32 +337,34 @@ class Table(object):
             # Keep the lock scope small because this is run very often.
             with self.db.lock:
                 self._threading_warn()
-                self._table = SQLATable(self.name,
-                                        self.db.metadata,
-                                        schema=self.db.schema)
+                self._table = SQLATable(
+                    self.name, self.db.metadata, schema=self.db.schema
+                )
                 if self._primary_id is not False:
                     # This can go wrong on DBMS like MySQL and SQLite where
                     # tables cannot have no columns.
                     primary_id = self._primary_id
                     primary_type = self._primary_type
                     increment = primary_type in [Types.integer, Types.bigint]
-                    column = Column(primary_id, primary_type,
-                                    primary_key=True,
-                                    autoincrement=increment)
+                    column = Column(
+                        primary_id,
+                        primary_type,
+                        primary_key=True,
+                        autoincrement=increment,
+                    )
                     self._table.append_column(column)
                 for column in columns:
                     if not column.name == self._primary_id:
                         self._table.append_column(column)
                 self._table.create(self.db.executable, checkfirst=True)
+                self._columns = None
         elif len(columns):
             with self.db.lock:
                 self._reflect_table()
                 self._threading_warn()
                 for column in columns:
                     if not self.has_column(column.name):
-                        self.db.op.add_column(self.name,
-                                              column,
-                                              self.db.schema)
+                        self.db.op.add_column(self.name, column, self.db.schema)
                 self._reflect_table()
 
     def _sync_columns(self, row, ensure, types=None):
@@ -398,31 +398,31 @@ class Table(object):
         return ensure
 
     def _generate_clause(self, column, op, value):
-        if op in ('like',):
+        if op in ("like",):
             return self.table.c[column].like(value)
-        if op in ('ilike',):
+        if op in ("ilike",):
             return self.table.c[column].ilike(value)
-        if op in ('>', 'gt'):
+        if op in (">", "gt"):
             return self.table.c[column] > value
-        if op in ('<', 'lt'):
+        if op in ("<", "lt"):
             return self.table.c[column] < value
-        if op in ('>=', 'gte'):
+        if op in (">=", "gte"):
             return self.table.c[column] >= value
-        if op in ('<=', 'lte'):
+        if op in ("<=", "lte"):
             return self.table.c[column] <= value
-        if op in ('=', '==', 'is'):
+        if op in ("=", "==", "is"):
             return self.table.c[column] == value
-        if op in ('!=', '<>', 'not'):
+        if op in ("!=", "<>", "not"):
             return self.table.c[column] != value
-        if op in ('in'):
+        if op in ("in"):
             return self.table.c[column].in_(value)
-        if op in ('between', '..'):
+        if op in ("between", ".."):
             start, end = value
             return self.table.c[column].between(start, end)
-        if op in ('startswith',):
-            return self.table.c[column].like('%' + value)
-        if op in ('endswith',):
-            return self.table.c[column].like(value + '%')
+        if op in ("startswith",):
+            return self.table.c[column].like("%" + value)
+        if op in ("endswith",):
+            return self.table.c[column].like(value + "%")
         return false()
 
     def _args_to_clause(self, args, clauses=()):
@@ -432,32 +432,31 @@ class Table(object):
             if not self.has_column(column):
                 clauses.append(false())
             elif isinstance(value, (list, tuple, set)):
-                clauses.append(self._generate_clause(column, 'in', value))
+                clauses.append(self._generate_clause(column, "in", value))
             elif isinstance(value, dict):
                 for op, op_value in value.items():
                     clauses.append(self._generate_clause(column, op, op_value))
             else:
-                clauses.append(self._generate_clause(column, '=', value))
+                clauses.append(self._generate_clause(column, "=", value))
         return and_(*clauses)
 
     def _args_to_order_by(self, order_by):
         orderings = []
-        for ordering in ensure_tuple(order_by):
+        for ordering in ensure_list(order_by):
             if ordering is None:
                 continue
-            column = ordering.lstrip('-')
+            column = ordering.lstrip("-")
             column = self._get_column_name(column)
             if not self.has_column(column):
                 continue
-            if ordering.startswith('-'):
+            if ordering.startswith("-"):
                 orderings.append(self.table.c[column].desc())
             else:
                 orderings.append(self.table.c[column].asc())
         return orderings
 
     def _keys_to_args(self, row, keys):
-        keys = ensure_tuple(keys)
-        keys = [self._get_column_name(k) for k in keys]
+        keys = [self._get_column_name(k) for k in ensure_list(keys)]
         row = row.copy()
         args = {k: row.pop(k, None) for k in keys}
         return args, row
@@ -503,7 +502,7 @@ class Table(object):
         ::
             table.drop_column('created_at')
         """
-        if self.db.engine.dialect.name == 'sqlite':
+        if self.db.engine.dialect.name == "sqlite":
             raise RuntimeError("SQLite does not support dropping columns.")
         name = self._get_column_name(name)
         with self.db.lock:
@@ -512,11 +511,7 @@ class Table(object):
                 return
 
             self._threading_warn()
-            self.db.op.drop_column(
-                self.table.name,
-                name,
-                self.table.schema
-            )
+            self.db.op.drop_column(self.table.name, name, self.table.schema)
             self._reflect_table()
 
     def drop(self):
@@ -529,8 +524,8 @@ class Table(object):
                 self._threading_warn()
                 self.table.drop(self.db.executable, checkfirst=True)
                 self._table = None
+                self._columns = None
                 self.db._tables.pop(self.name, None)
-                self._flush_metadata()
 
     def has_index(self, columns):
         """Check if an index exists to cover the given ``columns``."""
@@ -544,7 +539,7 @@ class Table(object):
                 return False
         indexes = self.db.inspect.get_indexes(self.name, schema=self.db.schema)
         for index in indexes:
-            if columns == set(index.get('column_names', [])):
+            if columns == set(index.get("column_names", [])):
                 self._indexes.append(columns)
                 return True
         return False
@@ -557,7 +552,7 @@ class Table(object):
 
             table.create_index(['name', 'country'])
         """
-        columns = [self._get_column_name(c) for c in ensure_tuple(columns)]
+        columns = [self._get_column_name(c) for c in ensure_list(columns)]
         with self.db.lock:
             if not self.exists:
                 raise DatasetException("Table has not been created yet.")
@@ -570,6 +565,17 @@ class Table(object):
                 self._threading_warn()
                 name = name or index_name(self.name, columns)
                 columns = [self.table.c[c] for c in columns]
+
+                # MySQL crashes out if you try to index very long text fields,
+                # apparently. This defines (a somewhat random) prefix that
+                # will be captured by the index, after which I assume the engine
+                # conducts a more linear scan:
+                mysql_length = {}
+                for col in columns:
+                    if isinstance(col.type, MYSQL_LENGTH_TYPES):
+                        mysql_length[col.name] = 10
+                kw["mysql_length"] = mysql_length
+
                 idx = Index(name, *columns, **kw)
                 idx.create(self.db.executable)
 
@@ -602,19 +608,17 @@ class Table(object):
         if not self.exists:
             return iter([])
 
-        _limit = kwargs.pop('_limit', None)
-        _offset = kwargs.pop('_offset', 0)
-        order_by = kwargs.pop('order_by', None)
-        _streamed = kwargs.pop('_streamed', False)
-        _step = kwargs.pop('_step', QUERY_STEP)
+        _limit = kwargs.pop("_limit", None)
+        _offset = kwargs.pop("_offset", 0)
+        order_by = kwargs.pop("order_by", None)
+        _streamed = kwargs.pop("_streamed", False)
+        _step = kwargs.pop("_step", QUERY_STEP)
         if _step is False or _step == 0:
             _step = None
 
         order_by = self._args_to_order_by(order_by)
         args = self._args_to_clause(kwargs, clauses=_clauses)
-        query = self.table.select(whereclause=args,
-                                  limit=_limit,
-                                  offset=_offset)
+        query = self.table.select(whereclause=args, limit=_limit, offset=_offset)
         if len(order_by):
             query = query.order_by(*order_by)
 
@@ -623,9 +627,7 @@ class Table(object):
             conn = self.db.engine.connect()
             conn = conn.execution_options(stream_results=True)
 
-        return ResultIter(conn.execute(query),
-                          row_type=self.db.row_type,
-                          step=_step)
+        return ResultIter(conn.execute(query), row_type=self.db.row_type, step=_step)
 
     def find_one(self, *args, **kwargs):
         """Get a single result from the table.
@@ -639,8 +641,8 @@ class Table(object):
         if not self.exists:
             return None
 
-        kwargs['_limit'] = 1
-        kwargs['_step'] = None
+        kwargs["_limit"] = 1
+        kwargs["_step"] = None
         resiter = self.find(*args, **kwargs)
         try:
             for row in resiter:
@@ -694,10 +696,12 @@ class Table(object):
         if not len(columns):
             return iter([])
 
-        q = expression.select(columns,
-                              distinct=True,
-                              whereclause=clause,
-                              order_by=[c.asc() for c in columns])
+        q = expression.select(
+            columns,
+            distinct=True,
+            whereclause=clause,
+            order_by=[c.asc() for c in columns],
+        )
         return self.db.query(q)
 
     # Legacy methods for running find queries.
@@ -717,4 +721,4 @@ class Table(object):
 
     def __repr__(self):
         """Get table representation."""
-        return '<Table(%s)>' % self.table.name
+        return "<Table(%s)>" % self.table.name
